import pandas as pd
import geopandas as gpd
from bs4 import BeautifulSoup, SoupStrainer
import zipfile
import wget
import os
import requests
from datetime import datetime
from zipfile import ZipFile
from urllib.request import urlopen 

def getData(State=None):
    dir_list = os.listdir()
    f = None
    state_ids = pd.read_csv(r'https://raw.githubusercontent.com/Fair-Lines-America/FLA_basic_tools/main/data/us-state-ansi-fips.csv',skipinitialspace=True, dtype=str)
    st = state_ids[state_ids['st']==State]
    st = st['stname'].iloc[0]
    if '2020_PLSummaryFile_FieldNames.xlsx' not in dir_list:
        headers='https://www2.census.gov/programs-surveys/decennial/rdo/about/2020-census-program/Phase3/SupportMaterials/2020_PLSummaryFile_FieldNames.xlsx'
        wget.download(headers)
    url = f'https://www2.census.gov/programs-surveys/decennial/2020/data/01-Redistricting_File--PL_94-171/{st}/'
    r = requests.get(url)
    res = r.content
    for link in BeautifulSoup(res, parse_only=SoupStrainer('a')):
        if 'zip' in link.contents[0]:
            f = link.contents[0]
    if f not in dir_list:
        file = f'{url}{f}'
        wget.download(file)
    if f'PL94_blocks_{State}.csv' not in dir_list:
        df_header1 = pd.read_excel('2020_PLSummaryFile_FieldNames.xlsx', sheet_name='2020 P.L. Segment 1 Definitions').dropna(axis=0, how='all').reset_index(drop=True)
        df_header2 = pd.read_excel('2020_PLSummaryFile_FieldNames.xlsx', sheet_name='2020 P.L. Segment 2 Definitions').dropna(axis=0, how='all').reset_index(drop=True)
        df_header3 = pd.read_excel('2020_PLSummaryFile_FieldNames.xlsx', sheet_name='2020 P.L. Segment 3 Definitions').dropna(axis=0, how='all').reset_index(drop=True)
        df_headergeo = pd.read_excel('2020_PLSummaryFile_FieldNames.xlsx', sheet_name='2020 P.L. Geoheader Definitions').dropna(axis=0, how='all').reset_index(drop=True)
        header_replace_1 = {i :None for i in range(0,len(df_header1.index)) }
        header_replace_2 = {i :None for i in range(0,len(df_header2.index)) }
        header_replace_3 = {i :None for i in range(0,len(df_header3.index)) }
        header_replace_geo = {i :None for i in range(0,len(df_headergeo.index)) }
        array = [[df_header1,header_replace_1, '1'],[df_header2,header_replace_2,'2'],[df_header3,header_replace_3,'3'],[df_headergeo,header_replace_geo,'o']]
        for i in array:
            json = i[1]
            header = i[0]
            for key in json.keys():
                json[key] = header.iloc[key][1]
        archive = zipfile.ZipFile(f, 'r')
        csv = []
        for i in archive.infolist():
            temp = archive.open(i)
            fileName = temp.name.split('.')[0]
            fileNum = fileName[-5:][0]
            df = pd.read_csv(temp, sep="|", header=None, low_memory=False ,encoding = "ISO-8859-1")
            for j in array:
                if fileNum == j[2] :
                    df = df.rename(columns=j[1])
            df.to_csv(f'{fileName}.csv', index=False)
            csv.append(fileName)
        join_on = ['STUSAB','LOGRECNO']
        df_out = None
        for i in csv:
            if df_out is None:
                df_out = pd.read_csv(f'{i}.csv', low_memory=False, dtype={'FILEID':'str','STUSAB':'str','CHARITER':'str','CIFSN':'str','LOGRECNO':'str'})
                continue
            else:
                df = pd.read_csv(f'{i}.csv', low_memory=False, dtype={'FILEID':'str','STUSAB':'str','CHARITER':'str','CIFSN':'str','LOGRECNO':'str'})
                df_out = df_out.merge(df, on=join_on, suffixes=('', '_y'))
                delt = []
                for k in df_out.columns:
                    if '_y' in k:

                        delt.append(k)
                df_out = df_out.drop(columns=delt)
        r = [i for i in df_out.columns if i[0:2] != 'P0' or i != 'GEOID20']
        df_out = df_out[df['SUMLEV'] == 750]
        
        keep = ['POP100',
        'CD116',
        'SLDU18',
        'SLDL18',
        'GEOCODE']
        r = [i for i in df_out.columns if i not in keep]
        df_out = df_out.drop(columns=r)
        df_out.to_csv(f'PL94_blocks_{State}.csv', index=False)
        for i in csv:
            os.remove(f'{i}.csv')
    else:
        df_out = pd.read_csv(f'PL94_blocks_{State}.csv', dtype={'GEOCODE':str,'CD116':str,'SLDU18':str,'SLDL18':str})
    return df_out

#####  PROCESS Off PL94 File  ######
def Overlap_old_new(new_districts, geoid='GEOID', district='District', leg=None):

#####  CHECKS  ######
    if leg not in ['CD116','SLDU18','SLDL18']:
        raise Exception('Please Choose Legislative level Default \'CD116\',\'SLDU18\',\'SLDL18\'')
    if not isinstance(new_districts ,pd.DataFrame):
        raise Exception('Please Supply a path to a CSV or DataFrame for new districts')
    if district not in new_districts.columns:
        raise Exception('District column not given or is missing from a DataFrame')
    if geoid not in new_districts.columns:
        raise Exception('GEOID column not given or is missing from a DataFrame')
    fips = new_districts[geoid].iloc[0]
    old_org = getData(State=str(fips[0:2]))
    old_org = old_org.rename(columns={'GEOCODE':geoid,leg:district})
    data = old_org[[geoid,'POP100']]
    old_districts = old_org[[geoid,district]]
#####  PROCESS  ######

    groupby_pop = district
    old = data.merge(old_districts, on=geoid)
    new = data.merge(new_districts, on=geoid)
    new = new.drop_duplicates()
    old = old.drop_duplicates()
    cross = old[[geoid,district]].merge(new,
                                        on=geoid,
                                        suffixes=('_old', '_new'))
    cross = cross.drop_duplicates()
    new = new.drop(columns=geoid)
    old = old.drop(columns=geoid)
    cross = cross.drop(columns=geoid)
    old = old.groupby([groupby_pop], as_index=False).sum()
    new = new.groupby([groupby_pop], as_index=False).sum()
    new_group = [district+'_old',district+'_new']
    cross_old, cross_new = [district+'_old',district+'_new']
    groupby_cross = district
    cross = cross.groupby([f'{groupby_cross}_old',f'{groupby_cross}_new'], as_index=False ).sum()
    new = cross.merge(new, left_on=cross_new, right_on=district,suffixes=('_new', '_cross'))
    old = cross.merge(old, left_on=cross_old, right_on=district,suffixes=('_old', '_cross'))
    out = new.merge(old, on=[f'{district}_new',f'{district}_old'] , suffixes=('', '_d'))
    r = [district]
    for col in out.columns:
        if '_d' in col:
            r.append(col)
    out = out.drop(columns=r)
    ## OUTPUT
    out.to_csv(f'out_population_overlap.csv',index=False)
    return out

#####  PROCESS Off 2 BAF ######
def Overlap_compare(old_districts, new_districts, data, geoid='GEOID', district='District'):

#####  CHECKS  ######
    if district not in old_districts.columns and district not in new_districts.columns:
        raise Exception('District column not given or is missing from a DataFrame')
    if geoid not in old_districts.columns and geoid not in new_districts.columns and geoid not in data.columns:
        raise Exception('GEOID column not given or is missing from a DataFrame')


#####  PROCESS ######

    groupby_pop = district
    old = data.merge(old_districts, on=geoid)
    new = data.merge(new_districts, on=geoid)
    new = new.drop_duplicates()
    old = old.drop_duplicates()
    cross = old[[geoid,district]].merge(new,
                                     on=geoid,
                                     suffixes=('_old', '_new'))
    cross = cross.drop_duplicates()
    new = new.drop(columns=geoid)
    old = old.drop(columns=geoid)
    cross = cross.drop(columns=geoid)
    old = old.groupby([groupby_pop], as_index=False).sum()
    new = new.groupby([groupby_pop], as_index=False).sum()
    new_group = [district+'_old',district+'_new']
    cross_old, cross_new = [district+'_old',district+'_new']
    groupby_cross = district
    cross = cross.groupby([f'{groupby_cross}_old',f'{groupby_cross}_new'], as_index=False ).sum()
    new = cross.merge(new, left_on=cross_new, right_on=district,suffixes=('_new', '_cross'))
    old = cross.merge(old, left_on=cross_old, right_on=district,suffixes=('_old', '_cross'))
    out = new.merge(old, on=[f'{district}_new',f'{district}_old'] , suffixes=('', '_d'))
    r = [district]
    for col in out.columns:
        if '_d' in col:
            r.append(col)
    out = out.drop(columns=r)
    ## OUTPUT
    out.to_csv(f'out_population_overlap.csv',index=False)
    return out


import pandas as pd
import geopandas as gpd
from bs4 import BeautifulSoup, SoupStrainer
import zipfile
import wget
import os
import requests
from datetime import datetime
import json
from zipfile import ZipFile
from urllib.request import urlopen 

place_ref = r'https://github.com/Fair-Lines-America/FLA_basic_tools/blob/main/data/place_ref.zip?raw=true'
vtd_ref = r'https://github.com/Fair-Lines-America/FLA_basic_tools/blob/main/data/vtd_ref.zip?raw=true'
mil_ref = r'https://raw.githubusercontent.com/Fair-Lines-America/FLA_basic_tools/main/data/mil_ref.csv'
########
#Helper Functions
########
def get_links(url):
    url_base = 'https://www2.census.gov/geo/tiger/TIGER2020/COUNTY/'
    res = requests.get(url)
    res = res.content
    array = []
    f= None 
    for link in BeautifulSoup(res, parse_only=SoupStrainer('a')):
        if link.has_attr('href') and '.zip'  in link['href']:
            array.append(link['href'])
    return array

########
#Split Functions
########


def community_split(distr, geoid, disid):
    if not isinstance(distr, pd.DataFrame):
        raise Exception('distr is not a Dataframe')
    url_base = 'https://www2.census.gov/geo/tiger/TIGER2020/'
    f = None 
    distr = distr[[disid,geoid]]
    distr = distr.astype(str)
    if len(distr[geoid].iloc[0]) < 15:
        raise Exception('GEOID column not in 15 digital Geocode for Census Blocks')
    state = distr[geoid].iloc[0][:2]
    links = get_links(f'{url_base}TABBLOCK20') 
    for l in links:
        if state in l.split('_')[2]:
            f = l
            break
    wget.download(f'{url_base}/TABBLOCK20/{f}')
    tabblock = gpd.read_file(f, dtype={'GEOID20': str})
    distr = distr.rename(columns={geoid:'GEOID20'})
    df_dist = tabblock.merge(distr, on='GEOID20')
    os.remove(f)
    if len(df_dist.index) == 0:
        raise Exception('Pandas Merge Fail')
    #County Split
    keep=[disid,'COUNTYFP20']
    remove=[]
    for i in df_dist.columns:
        if i not in keep:
            remove.append(i)
    df_dist = df_dist.drop(columns=remove)
    district_list = list(set(df_dist['COUNTYFP20'].values.tolist()))
    pivot_tbd = {}
    for i in district_list:
        pivot_tbd[i] = []
    for idx, row in df_dist.iterrows():
        pivot_tbd[row['COUNTYFP20']].append(row[disid])
        pivot_tbd[row['COUNTYFP20']] = list(set(pivot_tbd[row['COUNTYFP20']]))
    r = []
    for key, value in pivot_tbd.items():
        if len(value) < 2:
            r.append(pivot_tbd[key])
    Segement_Count =  0
    Split_Count = 0
    for key,value in pivot_tbd.items():
        if len(value) > 1:
            Segement_Count += len(value)
            Split_Count += 1
    out_county = {'Segement_Count':Segement_Count, 'Split_Count':Split_Count, 'County_List': pivot_tbd}
    r =[]
    for key, value in out_county['County_List'].items():
        if len(value) < 2:
            r.append(key)
    for i in r:
        del out_county['County_List'][i]
#### EXTERNAL FILES FOR REFERNCE NEEDED ####
    url = urlopen(place_ref)
    output = open('place_ref.zip', 'wb')    # note the flag:  "wb"        
    output.write(url.read())
    output.close()
    df_place = pd.read_csv('place_ref.zip', dtype=str)
    df_place = distr.merge(df_place, left_on='GEOID20', right_on='GEOID20_0KM', suffixes=('_d', '') )
    df_place = df_place.drop(columns=['GEOID20_d','GEOID20_0KM'])
    district_list = list(set(df_place['GEOID20'].values.tolist()))
    pivot_tbd = {}
    for i in district_list:
        pivot_tbd[i] = []
    for idx, row in df_place.iterrows():
        pivot_tbd[row['GEOID20']].append(row[disid])
        pivot_tbd[row['GEOID20']] = list(set(pivot_tbd[row['GEOID20']]))
    r = []
    for key, value in pivot_tbd.items():
        if len(value) < 2:
            r.append(pivot_tbd[key])
    Segement_Count =  0
    Split_Count = 0
    for key,value in pivot_tbd.items():
        if len(value) > 1:
            Segement_Count += len(value)
            Split_Count += 1
    out_place = {'Segement_Count':Segement_Count, 'Split_Count':Split_Count, 'Place_List': pivot_tbd}
    r =[]
    for key, value in out_place['Place_List'].items():
        if len(value) < 2:
            r.append(key)
    for i in r:
        del out_place['Place_List'][i]
    df_mil = pd.read_csv(mil_ref, dtype=str)
    df_mil = distr.merge(df_mil, left_on='GEOID20', right_on='GEOID20_0KM', suffixes=('', '') )
    df_mil = df_mil.drop(columns=['GEOID20','GEOID20_0KM'])
    district_list = list(set(df_mil['AREAID'].values.tolist()))
    pivot_tbd = {}
    for i in district_list:
        pivot_tbd[i] = []
    for idx, row in df_mil.iterrows():
        pivot_tbd[row['AREAID']].append(row[disid])
        pivot_tbd[row['AREAID']] = list(set(pivot_tbd[row['AREAID']]))
    r = []
    for key, value in pivot_tbd.items():
        if len(value) < 2:
            r.append(pivot_tbd[key])
    Segement_Count =  0
    Split_Count = 0
    for key,value in pivot_tbd.items():
        if len(value) > 1:
            Segement_Count += len(value)
            Split_Count += 1
    out_mil = {'Segement_Count':Segement_Count, 'Split_Count':Split_Count, 'Mil_List': pivot_tbd}
    r = []
    for key, value in out_mil['Mil_List'].items():
        if len(value) < 2:
            r.append(key)
    for i in r:
        del out_mil['Mil_List'][i]
    url = urlopen(vtd_ref)
    output = open('vtd_ref.zip', 'wb')    # note the flag:  "wb"        
    output.write(url.read())
    output.close()
    df_vtd = pd.read_csv('vtd_ref.zip', dtype=str)
    df_vtd = distr.merge(df_vtd, left_on='GEOID20', right_on='GEOID20', suffixes=('_d', '') )
    df_vtd = df_vtd.drop(columns=['GEOID20'])
    district_list = list(set(df_vtd['VTD_GEOID20'].values.tolist()))
    pivot_tbd = {}
    for i in district_list:
        pivot_tbd[i] = []
    for idx, row in df_vtd.iterrows():
        pivot_tbd[row['VTD_GEOID20']].append(row[disid])
        pivot_tbd[row['VTD_GEOID20']] = list(set(pivot_tbd[row['VTD_GEOID20']]))
    r = []
    for key, value in pivot_tbd.items():
        if len(value) < 2:
            r.append(pivot_tbd[key])
    Segement_Count =  0
    Split_Count = 0
    for key,value in pivot_tbd.items():
        if len(value) > 1:
            Segement_Count += len(value)
            Split_Count += 1
    out_vtd = {'Segement_Count':Segement_Count, 'Split_Count':Split_Count, 'Vtd_List': pivot_tbd}
    r = []
    for key, value in out_vtd['Vtd_List'].items():
        if len(value) < 2:
            r.append(key)
    for i in r:
        del out_vtd['Vtd_List'][i]
    out = {'vtd' : out_vtd,
        'county' : out_county,
        'place' : out_place,
        'mil' : out_mil}
    ## OUTPUT
    with open('geo_splits.json', 'w') as f:
        json.dump(out, f)
    return out


import click
import os
from .geo_splits import get_links,community_split
from .data_build import get_census_shp, assign_baf
from .overlap import getData, Overlap_old_new, Overlap_compare
from tabulate import tabulate
import pandas as pd
import geopandas as gpd
     
@click.command()
@click.option('--program', '-t', prompt='Choice Process', type=click.Choice(['assign_baf', 'community_split', 'overlap_compare']))
@click.option('--path',prompt='Path to working Directory', default=os.getcwd(), type=click.Path(exists=True))
def main(path, program):
    os.chdir(path)
    if program == 'assign_baf':
        df = click.prompt('Please enter File', type=str)
        if 'csv' in df.split('.'):
            csv = True
            table_sample = pd.read_csv(f'{df}',nrows=2)
            table = pd.read_csv(f'{df}',dtype=str)
        elif 'shp' in df.split('.'):
            table_sample = gpd.read_file(f'{df}',nrows=2)
            table = gpd.read_file(f'{df}',dtype=str)
        else:
            raise Exception("Please Supply a CSV or SHP file")
        click.echo(tabulate(table_sample, headers='keys', tablefmt='psql',showindex=False))
        click.confirm('Do you want to continue?', abort=True)
        district = click.prompt('Please enter Name of District Column', type=str)
        if csv:
            geoid = click.prompt('Please enter Name of 15 digit GEOID', type=str)
        else:
            geoid=None
        state = click.prompt('Please enter State for the Poltical Geography', type=str)
        assign_baf(table, state, district, geoid)
    if program == 'community_split':
        df = click.prompt('Please enter CSV', type=str)
        table_sample = pd.read_csv(f'{df}',nrows=2)
        table = pd.read_csv(f'{df}',dtype=str)
        click.echo(tabulate(table_sample, headers='keys', tablefmt='psql',showindex=False))
        click.confirm('Do you want to continue?', abort=True)
        district = click.prompt('Please enter Name of District Column', type=str)
        geoid = click.prompt('Please enter Name of 15 digit GEOID', type=str)
        community_split(distr=table, geoid=geoid, disid=district)
    if program == 'community_split':
        df = click.prompt('Please enter CSV', type=str)
        table_sample = pd.read_csv(f'{df}',nrows=2)
        table = pd.read_csv(f'{df}',dtype=str)
        click.echo(tabulate(table_sample, headers='keys', tablefmt='psql',showindex=False))
        click.confirm('Do you want to continue?', abort=True)
        district = click.prompt('Please enter Name of District Column', type=str)
        geoid = click.prompt('Please enter Name of 15 digit GEOID', type=str)
        leg = click.prompt('Please Choice Leg from List', type=click.Choice(['CD116', 'SLDU18', 'SLDL18']))
        Overlap_old_new(new_districts=table, geoid=geoid, district=district, leg=leg)
if __name__ == '__main__':
    main()